# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import copy
import json
import logging

import io
import pprint
import textwrap
import time

from . import entity_json
from .exceptions import ConfigUploadFailed, ExceptionWithResponse
from .entitybase import EntityBase
from .pump import Pump

logger = logging.getLogger(__name__)


class Pipe(EntityBase):
    """
    This class represents a pipe.
    """

    def __str__(self):
        result = super().__str__()
        # TODO: add info about task, provider and sink.
        return result

    def get_pump(self, **kwargs):
        url = self._connection.get_pipe_pump_url(self.id)
        response = self._connection.do_get_request(url, allowable_response_status_codes=[200], **kwargs)
        pump_json = response.json()
        return Pump(self, self._connection, pump_json)

    @property
    def source_dataset_id(self):
        """Returns the id of the dataset that this pipe reads data from, or None if the pipe doesn't
        read data from a dataset."""
        effective_config = self._raw_jsondata["config"]["effective"]
        producer_config = effective_config.get("source")
        if producer_config is None:
            return None

        return producer_config.get("dataset")

    @property
    def sink_dataset_id(self):
        """Returns the id of the dataset that this pipe feeds data to, or None if the pipe doesn't
        feed data to a dataset."""
        effective_config = self._raw_jsondata["config"]["effective"]
        sink_config = effective_config.get("sink")
        if sink_config is None:
            return None

        return sink_config.get("dataset")

    @property
    def execution_dataset_id(self):
        """Returns the id of the dataset that this pipe logs to, or None if the pipe doesn't
        log to a dataset (this is the case for the pipes of the old sdshare server and client)."""
        runtime = self._raw_jsondata["runtime"]
        return runtime.get("execution-dataset")

    def get_log_entities(self, since=None, limit=None, start=None, history=True, deleted=True, reverse=False,
                         do_transit_decoding=True, accept="application/json", expected_node_id=None, method="GET",
                         **kwargs):
        """Returns the content of the execution dataset using the /pipes/{pipe}/log API """
        url = self._connection.get_pipe_log_url(self.id)

        params = {}
        if since is not None:
            params["since"] = since
        if limit is not None:
            params["limit"] = limit
        if start is not None:
            params["start"] = start
        if history is not None:
            params["history"] = history
        if deleted is not None:
            params["deleted"] = deleted
        if reverse is not None:
            params["reverse"] = reverse

        headers = {"ACCEPT": accept}
        if expected_node_id is not None:
            headers["X-Expected-Node-Id"] = expected_node_id

        # The entities are streamed from the server, so we have to enable streaming here.
        request_func = {
            "GET": self._connection.do_get_request,
            "OPTIONS": self._connection.do_options_request
        }[method]

        if request_func is None:
            raise AssertionError("Unsupported request-method '%s'!" % (method,))

        response = request_func(url, allowable_response_status_codes=[200], stream=True, params=params,
                                headers=headers, **kwargs)

        return entity_json.parse_json_stream(response.raw, do_transit_decoding=do_transit_decoding)

    @property
    def runtime(self):
        return copy.deepcopy(self._raw_jsondata["runtime"])

    @property
    def graph(self):
        return copy.deepcopy(self._raw_jsondata["graph"])

    def get_entities(self, since=None, limit=None, stage=None, do_transit_decoding=True, **kwargs):
        """This returns a generator for iterating over the entities of the specified stage
        of this pipe.
        :param do_transit_decoding: bool determining if transit decoding should be performed
        :param since:
        :param limit: The maximum number of entities to return. A value of None means return all the entities.
        :param stage:
        """
        url = self._connection.get_pipe_entities_url(self.id)

        params = {}
        if since is not None:
            params["since"] = since
        if limit is not None:
            params["limit"] = limit
        if stage is not None:
            params["stage"] = stage

        # The entities are streamed from the server, so we have to enable streaming here.
        response = self._connection.do_get_request(url, allowable_response_status_codes=[200],
                                                   stream=True, params=params, **kwargs)

        return entity_json.parse_json_stream(response.raw, do_transit_decoding=do_transit_decoding)

    def post_entities(self, entities,
                      is_full=None,
                      sequence_id=None,
                      request_id=None,
                      previous_request_id=None,
                      is_last=None,
                      is_first=None,
                      completeness=None,
                      headers=None,
                      **kwargs
                      ):
        """This posts the specified entities to the pipe. This only works for pipes that has a source
        of the type 'http_endpoint'.
        """
        url = self._connection.get_pipe_receiver_endpoint_url(self.id)

        # If the size of the entities is very large, we have to use chunked encoding; otherwise we will get
        # a "connection reset" error.
        if type(entities) is str:
            # The entities were already json-encoded.
            entities_as_json = entities
        else:
            entities_as_json = entity_json.entities_to_json(entities)

        class BytesIOWithNoLength(io.BytesIO):
            def __len__(self):
                raise io.UnsupportedOperation()

            def __bool__(self):
                return True

        entities_stream = BytesIOWithNoLength(entities_as_json.encode("utf-8"))

        params = {}
        if is_full is not None:
            params["is_full"] = True
        if sequence_id is not None:
            params["sequence_id"] = sequence_id
        if request_id is not None:
            params["request_id"] = request_id
        if previous_request_id is not None:
            params["previous_request_id"] = previous_request_id
        if is_last is not None:
            params["is_last"] = is_last
        if is_first is not None:
            params["is_first"] = is_first

        if headers is None:
            headers = {"Content-Type": "application/json"}
        if completeness is not None:
            headers["X-Dataset-Completeness"] = str(completeness)
        response = self._connection.do_post_request(url,
                                                    allowable_response_status_codes=[200],
                                                    headers=headers,
                                                    params=params,
                                                    data=entities_stream, **kwargs)

        return response.json()

    def delete(self, **kwargs):
        """Deletes this pipe from the sesam-node.
        """
        url = self._connection.get_pipe_url(self.id)
        response = self._connection.do_delete_request(url, allowable_response_status_codes=[200], **kwargs)

        self.update_raw_jsondata(response.content.decode("utf-8"))

    def modify(self, pipe_config, force=False, **kwargs):
        """Modifies the pipe with the specified configuration."""
        request_params = {"force": "true" if force else "false"}
        pipe_config = copy.deepcopy(pipe_config)
        response = self._connection.do_put_request(self._connection.get_pipe_config_url(self.id),
                                                   allowable_response_status_codes=[200, 400],
                                                   json=pipe_config,
                                                   params=request_params, **kwargs)

        if response.status_code == 400:
            raise ConfigUploadFailed(response=response)

        self.update_raw_jsondata(response.content.decode("utf-8"))

    def wait_for_pipe_to_be_deployed(self, timeout=300, **kwargs):
        starttime = time.monotonic()
        while True:
            try:
                self.update_raw_jsondata_from_url(self._connection.get_pipe_url(self.id), **kwargs)
            except ExceptionWithResponse as e:
                if e.response.status_code == 404:
                    # This pipe has been deleted according to the API, we can return
                    return
                elapsetime = time.monotonic() - starttime
                if elapsetime > timeout:
                    raise AssertionError(
                        f"Timed out while waiting for the config of the pipe '{self.id}' to be deployed. "
                        f"exception: {e}")
                time.sleep(1)
                continue

            runtime = self.runtime
            if "state" not in self.runtime:
                raise AssertionError(f"No 'state' property found in self.runtime! self.runtime: {runtime}")
            if self.runtime["state"] != "Deploying":
                try:
                    self.get_pump()
                    return
                except:
                    logger.warning(f"The pipe with id '{self.id}' has been deployed, but an error was raised while "
                                   f"accessing the pump. Will retry.")
                    pass
            elapsetime = time.monotonic() - starttime
            if elapsetime > timeout:
                raise AssertionError(
                    f"Timed out while waiting for the config of the pipe '{self.id}' to be deployed. "
                    f"self._raw_jsondata: {textwrap.indent(pprint.pformat(self._raw_jsondata), prefix='    ')}")
            time.sleep(1)

    def preview(self, pipe_config, source_entity, do_transit_decoding=True):
        response = self._connection.do_post_request(self._connection.get_pipe_url(self.id) + "/preview",
                                                    allowable_response_status_codes=[200],
                                                    data={
                                                        "pipe-config": json.dumps(pipe_config),
                                                        "source": json.dumps(source_entity)},
                                                    stream=True)
        result = list(entity_json.parse_json_stream(response.raw, do_transit_decoding=do_transit_decoding))
        assert len(result) == 1
        result = result[0]
        return result



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import copy
import time
import json
import logging

from . import entity_json

from .exceptions import TimeoutWhileWaitingForRunningIndexToFinishUpdatingException
from .entitybase import EntityBase
from .utils import validate_response_is_ok

logger = logging.getLogger(__name__)


class Dataset(EntityBase):
    """
    This class represents a dataset.
    """

    def __str__(self):
        result = super().__str__()
        return result

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_entities_as_response(self, since=None, limit=None, start=None, history=True, deleted=True,
                                 accept="application/json", reverse=False, uncommitted=None,
                                 subset=None, expected_node_id=None,
                                 method="GET", **kwargs
                                 ):
        """
        :param since: The opaque value that can be used to efficiently skip already seen entities.

        :param limit: The "limit" parameter specifies the maximum number of entities to return.
                      If this is not specified, all available entities are returned.

        :param start: The "start" parameter is a positive integer that specified how many entities to skip from the
                      start of where the "since"-parameters starts. This is far less efficient than using the
                      "since"-parameter, but it can be manually set in a gui to explore the dataset. NOTE: Clients
                      should use the "since"-parameter whenever possible.

        :param history: If this is true (the default) all entities, including replaced ones, will be returned. \
                        If this is false, only the latest version of the entities will be returned.

        :param deleted: If this is True (the default) deleted entities will also be returned. \
                        If this is False, only the non-deleted entities will be returned.

        :param accept:  The value to use in the "Accept" header

        :param reverse: If this is False (the default) entities will returned with the oldest entity first. \
                        If this is True, the newest entity will be returned first.

        :param uncommitted: If true then the response may include uncommitted entities. This is only relevant if
                        the pipe that writes to the dataset have a circuit breaker enabled. Use this for debugging
                        purposes only.

        :param subset: If specified then the specified subset expression will be used to retrive entities in
                       the dataset via a secondary index. If the index does not exist, then 404 is returned.

        :param expected_node_id: If specified then it will be sent in the the X-Expected-Node-Id request header.
        """
        url = self._connection.get_dataset_entities_url(self.id)

        params = {}
        if since is not None:
            params["since"] = since
        if limit is not None:
            params["limit"] = limit
        if start is not None:
            params["start"] = start
        if history is not None:
            params["history"] = history
        if deleted is not None:
            params["deleted"] = deleted
        if reverse is not None:
            params["reverse"] = reverse
        if uncommitted is not None:
            params["uncommitted"] = uncommitted
        if subset is not None:
            params["subset"] = subset

        headers = {"ACCEPT": accept}
        if expected_node_id is not None:
            headers["X-Expected-Node-Id"] = expected_node_id

        # The entities are streamed from the server, so we have to enable streaming here.
        request_func = {
            "GET": self._connection.do_get_request,
            "OPTIONS": self._connection.do_options_request,
            "HEAD": self._connection.do_head_request,
        }[method]
        if request_func is None:
            raise AssertionError("Unsupported request-method '%s'!" % (method,))

        response = request_func(url, allowable_response_status_codes=[200], stream=True,
                                params=params, headers=headers, **kwargs)

        return response

    def get_entities_as_stream(self, since=None, limit=None, start=None, history=True, deleted=True,
                               accept="application/json", reverse=False, uncommitted=None,
                               subset=None, expected_node_id=None, **kwargs
                               ):
        """This returns a http stream for getting the entities of this dataset. (The returned object is a file-like
        stream from the 'requests' library, i.e response.raw)
        """
        response = self.get_entities_as_response(since=since, limit=limit, start=start, history=history,
                                                 deleted=deleted, accept=accept, reverse=reverse,
                                                 uncommitted=uncommitted,
                                                 subset=subset, expected_node_id=expected_node_id, **kwargs
                                                 )

        return response.raw

    def get_entities(self, do_transit_decoding=True, since=None, limit=None, start=None, history=True, deleted=True, reverse=False,
                     uncommitted=None,
                     subset=None, expected_node_id=None, **kwargs
                     ):
        """This returns a generator for iterating over the entities of this dataset.

        Note: If you just need the raw bytes, consider using the get_entities_as_stream() method instead. That will be
        faster.
        """
        entities_stream = self.get_entities_as_stream(since=since, limit=limit, start=start, history=history,
                                                      deleted=deleted, reverse=reverse, uncommitted=uncommitted,
                                                      subset=subset, expected_node_id=expected_node_id, **kwargs
                                                      )
        return entity_json.parse_json_stream(entities_stream, do_transit_decoding=do_transit_decoding)

    def get_entity(self, entity_id, do_transit_decoding=True, offset=None, uncommitted=None, **kwargs):
        """Returns the specified entity if it exist, or None if the entity does not exist.

        :param entity_id:
        """
        url = self._connection.get_dataset_entity_url(self.id)
        params = {"entity_id": entity_id}
        if offset is not None:
            params["offset"] = offset
        if uncommitted is not None:
            params["uncommitted"] = uncommitted

        response = self._connection.do_get_request(url,
                                                   allowable_response_status_codes=[200, 404],
                                                   params=params, **kwargs)
        if response.status_code == 404:
            # No such element.
            return None
        return entity_json.parse_entity(response.content.decode("utf-8"), do_transit_decoding=do_transit_decoding)

    def post_entities(self, entities, force=None, **kwargs):
        """This posts the specified entities to the dataset."""
        url = self._connection.get_dataset_entities_url(self.id)
        params = {}

        if force is not None:
            params["force"] = force

        response = self._connection.do_post_request(url,
                                                    allowable_response_status_codes=[200],
                                                    headers={"Content-Type": "application/json"},
                                                    data=entity_json.entities_to_json(entities),
                                                    params=params, **kwargs)
        return response.json()

    @property
    def populated(self):
        return self._raw_jsondata["runtime"].get("populated") is True

    @property
    def num_entities_in_index(self):
        return self._raw_jsondata["runtime"]["count-index-exists"]

    @property
    def num_deleted_entities_in_index(self):
        return self._raw_jsondata["runtime"]["count-index-deleted"]

    @property
    def num_undeleted_entities_in_index(self):
            return self.num_entities_in_index - self.num_deleted_entities_in_index

    @property
    def num_entities_in_log(self):
        return self._raw_jsondata["runtime"]["count-log-exists"]

    @property
    def num_deleted_entities_in_log(self):
        return self._raw_jsondata["runtime"]["count-index-deleted"]

    def get_indexes(self, **kwargs):
        url = self._connection.get_dataset_indexes_url(self.id)
        response = self._connection.do_get_request(url, **kwargs)
        indexes = response.json()
        return indexes

    def wait_for_indexes_to_finish_updating(self, timeout=60, verbosity=0, **kwargs):
        starttime = time.monotonic()
        time_of_last_log_output = 0
        while True:
            indexes = self.get_indexes(**kwargs)
            for index in indexes:
                if index["status"] not in ["up-to-date", "orphaned"]:
                    # at least one index is still being updated
                    if verbosity > 0:
                        current_time = time.monotonic()
                        if (time.monotonic() - time_of_last_log_output) > 1:
                            print("The index '%s' on the dataset '%s' has status '%s' and has "
                                  "indexed %d/%d entities." % (
                                      index["expression"],
                                      index["dataset"],
                                      index["status"],
                                      index["progress"]["indexed"],
                                      index["progress"]["total"],
                                  ))
                            time_of_last_log_output = current_time

                    break
            else:
                # all indexes are done
                return

            elapsedtime = time.monotonic() - starttime
            if elapsedtime > timeout:
                raise TimeoutWhileWaitingForRunningIndexToFinishUpdatingException(
                    "Timed out while waiting for the index to finish updating! dataset:\n%s" % (
                        self.as_json(),))
            time.sleep(0.5)

    def delete(self, **kwargs):
        """Deletes this dataset from the sesam-node.
        """
        url = self._connection.get_dataset_url(self.id)
        response = self._connection.do_delete_request(url, allowable_response_status_codes=[200], **kwargs)
        self.update_raw_jsondata(response.content.decode("utf-8"))

    def run_operation(self,
                      operation,
                      operation_parameters=None,
                      allowable_response_status_codes=frozenset([200]), **kwargs):
        """Runs the specified operation on the dataset.
        :param operation: The operation to run.
        :param operation_parameters: A dict with operation parameters.
        :param allowable_response_status_codes: The http request status code(s) to accept (defaults to 200).
        """
        url = self._connection.get_dataset_url(self.id)
        postdata = {"operation": operation}

        if operation_parameters is not None:
            postdata.update(operation_parameters)

        response = self._connection.do_post_request(url,
                                                    allowable_response_status_codes=allowable_response_status_codes,
                                                    data=postdata, **kwargs)

        return response.json()

    @property
    def runtime(self):
        return copy.deepcopy(self._raw_jsondata["runtime"])



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import os

from setuptools import setup, find_packages

requires = [
    "idna",
    "appdirs",
    "ijson",
    "python-dateutil",
    "pyyaml",
    "requests",
    "pytest",
    "requests_oauthlib",
    "oauthlib",
]

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.txt')) as f:
    README = f.read()

with open(os.path.join(here, "sesamclient", 'VERSION.txt')) as f:
    VERSION = f.read().strip()

setup(name='sesamclient',
      version=VERSION,
      description='sesamapi client',
      long_description=README,
      classifiers=[
          "Programming Language :: Python",
          "Topic :: Internet :: WWW/HTTP",
      ],
      author='Sesam',
      url='http://sesam.io',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      setup_requires=['pip'],
      test_suite="sesamclient/tests",
      tests_require=['pytest'],
      install_requires=requires,
      entry_points={
          'console_scripts': [
               'sesam=sesamclient.main.main:main',
          ],
      })



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import inspect
import os.path

from . import exceptions


def validate_in(key, mapping):
    if key not in mapping:
        raise AssertionError("The key '%s' was not found in the mapping %s!" % (key, mapping))


def validate_response_is_ok(response, allowable_response_status_codes=frozenset([200])):
    if isinstance(allowable_response_status_codes, int):
        allowable_response_status_codes = {allowable_response_status_codes}
    if response.status_code not in allowable_response_status_codes:
        if response.status_code == 403:
            exception_class_to_use = exceptions.ForbiddenException

        elif 400 <= response.status_code <= 499:
            exception_class_to_use = exceptions.BadRequestException

        elif response.status_code == 503:
            exception_class_to_use = exceptions.ServiceUnavailable

        else:
            exception_class_to_use = exceptions.InternalServerError

        try:
            response_text = response.content.decode("utf-8")
        except:
            response_text = response.text

        raise exception_class_to_use(
            """response.status_code(%s) not in allowable_response_status_codes(%s) for the url\
             '%s' (request method:'%s'). response.content:\n%s""" % (
                response.status_code, allowable_response_status_codes,
                response.url,
                response.request.method,
                response_text[:1000]),
            response=response
        )

    if is_called_by_unittest():
        # The sesamclient is being used by a ci-test, so do some additional checks.
        # check for headers with duplicate values

        # We must skip a couple of special case nodes that runs an old version of the code.
        url = response.request.url
        if not ("_msgpack-version-1" in url or "_msgpack-version-2" in url):
            for header_name, header_value in response.headers.items():
                if header_value:
                    header_values = header_value.split(",")
                    header_values_set = set()
                    for value in header_values:
                        value = value.strip()
                        if value in header_values_set:
                            raise AssertionError(
                                "Got a duplicate value '%s' in the '%s' response header for the url '%s'! "
                                "Header value: '%s'" % (
                                    value, header_name, response.request.url, header_value))
                        header_values_set.add(value)


def parse_json_response(response):
    """Utility-function for parsing a http response as json and reporting any parsing-errors in a human-readable
    way. The response.json() method doesn't give very helpful error-messages.

    :param response: The http response object
    :return: The response.content parsed as json
    """
    try:
        return response.json()
    except Exception as e:
        try:
            response_text = response.content.decode("utf-8")
        except:
            response_text = response.text

        raise AssertionError(
            """Failed to parse the response as json for the url\
             '%s' (request method:'%s'): %s. response.content:\n%s""" % (
                response.url,
                response.request.method,
                e,
                response_text[:1000]))


def validate_equal_case_insensitive(expected, actual, msg=None):
    if isinstance(expected, str) and isinstance(actual, str):
        if expected.lower() == expected.lower():
            # the arguments are strings and they are the same when ignoring case.
            return
    if msg:
        raise AssertionError("""actual(%s) != expected(%s)! %s""" % (actual, expected, msg))
    else:
        raise AssertionError("""actual(%s) != expected(%s)""" % (actual, expected))


def get_version():
    with open(os.path.join(os.path.dirname(__file__), 'VERSION.txt')) as version_file:
        return version_file.read().strip()


def is_called_by_unittest():
    stack = inspect.stack()
    if any(x[0].f_globals['__name__'].startswith('nose.') for x in stack):
        return True

    return any(x[0].f_globals['__name__'].startswith('_pytest') for x in stack)



# Copyright (C) Bouvet ASA - All Rights Reserved.
# Unauthorized copying of this file, via any medium is strictly prohibited.
import logging

from . import utils
from .entitybase import EntityBase

logger = logging.getLogger(__name__)


class Log(EntityBase):
    """
    This class represents a logfile.
    """

    def __str__(self):
        result = super().__str__()
        return result

    def get_content(self):
        """This returns a stream with the actual logfile content"""
        return self._connection.get_log_content(self.id)
